/*
 * RTEMS/KA9Q driver for NE2000 network card in SBC5206 ISA slot
 *
 * Written by David Fiddes, D.J.Fiddes@hw.ac.uk
 * http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  $Id:  $
 */
#include <bsp.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdarg.h>
#include <rtems/error.h>
#include <rtems/rtems_bsdnet.h>

#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/sockio.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <ns8390.h>


/* 0 = no debug, 1 = standard traces, 2 = register accesses, 3 = all accesses inc data */

#define DEBUG_NONE		0
#define DEBUG_STD		1
#define DEBUG_REG		2
#define	DEBUG_DATA		3

static const int debug_level = DEBUG_NONE;

static const char *version = 
	"NE2000 RTEMS/FreeBSD driver A0.01 22/08/98 David Fiddes (D.J@fiddes.surfaid.org)\n";


#if 0
  #define ISABASE (0x40000000)
#else
  #define ISABASE (0xFE600000)
#endif
#define WIERDOFFSET (0x10000)


	
/*
 * Module Prototypes
 */


/*
 * I/O function implementation
 */

static unsigned8
sbc5206_ne2000_getRegister( unsigned32 port, unsigned8 reg )
{
	volatile unsigned16 *pRegister;
	int					odd_addr;

	odd_addr = reg & 1;

	if ( odd_addr )
	{
		pRegister = (unsigned16 *)( ISABASE + WIERDOFFSET + port + reg - 1 );
	}
	else
	{
		pRegister = (unsigned16 *)( ISABASE + port + reg );
	}

	if (debug_level >= DEBUG_REG)
	{
		printf("sbc5206_ne2000_getRegister( port = %4.4X, reg = %X )\n", port, reg );
		printf("  pRegsiter = %4.4X\n", (unsigned32)pRegister );
	}

	return *pRegister;
}


static void
sbc5206_ne2000_setRegister(unsigned32 port, unsigned8 reg, unsigned8 value)
{
	volatile unsigned16 *pRegister;
	int					odd_addr;

	odd_addr = reg & 1;

	if ( odd_addr )
	{
		pRegister = (unsigned16 *)( ISABASE + WIERDOFFSET + port + reg - 1 );
	}
	else
	{
		pRegister = (unsigned16 *)( ISABASE + port + reg );
	}

	if (debug_level >= DEBUG_REG)
	{
		printf("ne2000_setRegister( port = %4.4X, reg = %X, value = %X )\n", port, reg, value );
		printf("  pRegister = %4.4X\n", (unsigned32)pRegister );
	}

	*pRegister = value;
}


static unsigned16
sbc5206_ne2000_getData( unsigned32 port )
{
	volatile unsigned16 *pRegister;

	pRegister = (unsigned16 *)( ISABASE + port + NS8390_PAGE0_DATAPORT );

	if (debug_level >= DEBUG_REG)
	{
	    printf("sbc5206_ne2000_getData( port = %4.4X )\n", port );
		printf("  pRegister = %4.4X\n", (unsigned32)pRegister );
	}

	return *pRegister;
}


static void
sbc5206_ne2000_setData(unsigned32 port, unsigned16 value)
{
	volatile unsigned16 *pRegister;

	pRegister = (unsigned16 *)( ISABASE + port + NS8390_PAGE0_DATAPORT );

	if (debug_level >= DEBUG_DATA)
	{
		printf("sbc5206_ne2000_setData( port = %4.4X, value = %2.2X )\n", port, value );
		printf("  pRegister = %4.4X\n", (unsigned32)pRegister );
	}

	*pRegister = value;
}


static void
ne2000_block_output( struct ns8390_device *dp, int *len, struct mbuf *packet_buf_start, int start_page )
{
	unsigned8				*p;
	int						index;

	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	setData_f				setData;

	struct mbuf				*bp;

	if (debug_level >= DEBUG_STD)
		printf( "ne2000_block_output() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	setData		= dp->setData;


	bp = packet_buf_start;

	/* flip through the mbufs describing the packet to work out how big it is */
	*len = bp->m_len;
	while ( bp->m_next != NULL )
	{
		bp = bp->m_next;
		*len = *len + bp->m_len;
	}


	/* since we're a word output device upsize to an even number of bytes */
	if (*len & 0x01)
		*len = *len + 1;
	

	/* Make sure we're on PAGE0, abort any transfer and stay started */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );


	/* set the size of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, *len & 0xFF );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, (*len >> 8) & 0xFF );

	/* set the source of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, 0x00 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR1, start_page );

	/* start remote write */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_WR );

	if (debug_level >= DEBUG_STD)
		printf( "About to spool the packet to the chip...\n" );

	/* spool each mbuf to the chip */
	bp = packet_buf_start;
	do {
		/* output the actual data */
		p = mtod( bp, unsigned8 * );
		for (index = 0; index < bp->m_len; index = index + 2)
		{
			(*setData)( ne2000_port, *(unsigned16*)(p + index) );
		}
		
		bp = bp->m_next;
	} while ( bp != NULL ); 

	if (debug_level >= DEBUG_STD)
		printf( "ne2000_block_output() - end\n" );
}

static void
ne2000_block_input( struct ns8390_device *dp, int len, struct mbuf *bp, int ring_offset )
{
	unsigned8				*p;
	int						index;

	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;

	if (debug_level >= DEBUG_STD)
		printf( "ne2000_block_input() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;

	/* Make sure we're on PAGE0, abort any transfer and stay started */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );

	/* set the size of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, len & 0xFF );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, (len >> 8) & 0xFF );

	/* set the source of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, ring_offset & 0xFF );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR1, (ring_offset >> 8) & 0xFF );

	/* start remote read */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_RD );

	bp->m_len = len;

	p = mtod( bp, unsigned8 * );
	for (index = 0; index < len; index = index + 2 )
	{
		*(unsigned16*)(p + index) = (*getData)( ne2000_port );
	}

	if (debug_level >= DEBUG_STD)
		printf( "ne2000_block_input - end\n" );
}


static void
ne2000_get_pkt_hdr( struct ns8390_device *dp, struct ns8390_pkt_hdr *header, int ring_page )
{
	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;
	unsigned16				a;
	unsigned16				b;

	if (debug_level >= DEBUG_STD)
		printf( "ne2000_get_pkt_hdr() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;

	/* Make sure we're on PAGE0, abort any transfer and stay started */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );

	/* set the size of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 4 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

	/* set the source of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR1, ring_page );

	/* start remote read */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_RD );


	a = (*getData)( ne2000_port );
	b = (*getData)( ne2000_port );


	header->status = (unsigned8)a;
	header->next   = a >> 8;
	header->count  = (b<<8)|(b>>8);

	
	if (debug_level >= DEBUG_STD)
		printf( "ne2000_get_pkt_hdr() : status = %.2X, next = %.2X, count = %.4X\n", header->status, header->next, header->count );
}


static void
ne2000_get_eth_addr( struct ns8390_device *dp, unsigned8 *hwa )
{
#if 0
	int						index;
	unsigned8				eeprom[32];
	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;


	if (debug_level >= DEBUG_STD)
		printf( "ne2000_get_eth_addr() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;


	/* read the ethernet address from eprom ? */

	/* Make sure we are on PAGE0, abort any current DMA transfers and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STP );


	/* set FIFO Threshold to 8bytes */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_DCR, NS8390_DCR_FT_8B );

	/* Reset the dma transfer size */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

	/* mask and clear all the chip's interrupts */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_IMR, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

	/* Set to do an internal loopback *//* MON seems undocumented */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_MON );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TCR, NS8390_TCR_LOOP_I0 );

	/* set transfer size to 32 bytes(sizeof EEPROM) */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 32 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );
	
	/* make transfer go to 0 */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR1, 0 );

	/* start the transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_RD );

	/* load in the EEPROM data */
	/* NE2000s double up the PROM data which is why this works */
	for (index = 0; index < 16; ++index)
		eeprom[index] = (*getData)( ne2000_port );

	/* suck out the Ethernet address that we need */
	for (index = 0; index < 6; index++)
		hwa[index] = eeprom[index*2]; 

	/* Abort any transfer and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STP | NS8390_CR_RD_AB );

	if (debug_level >= DEBUG_STD)
		printf( "ne2000_get_eth_addr() - end\n" );
#else
	hwa[0] = 0x00;
	hwa[1] = 0x10;
	hwa[2] = 0x8B;
	hwa[3] = 0xF1;
	hwa[4] = 0x13;
	hwa[5] = 0x00;
#endif
}



/*
 * Attach an NE2000 driver to the system
 */
int
rtems_ne2000_driver_attach (struct rtems_bsdnet_ifconfig *config)
{
	rtems_status_code		sc;
	rtems_isr_entry			old_handler;
 	struct ns8390_device	*dp;
	int						status;

	/* Print module version information */
	printf( version );


	/* Get a unique device pointer for our card */
	dp  = ns8390_get_free_device();

	if ( dp == NULL )
	{
		return 0;
	}

	/* Configure the NE2000 ns8390 configuration */

	/* Setup the device specific I/O functions */
	dp->getRegister = &sbc5206_ne2000_getRegister;
	dp->setRegister = &sbc5206_ne2000_setRegister;
	dp->getData	    = &sbc5206_ne2000_getData;
	dp->setData     = &sbc5206_ne2000_setData;


	/* Define the NE2000 NS8390 "personality" routines */

	dp->block_output   = &ne2000_block_output;
	dp->block_input    = &ne2000_block_input;
	dp->get_ns8390_hdr = &ne2000_get_pkt_hdr;
	dp->get_eth_addr   = &ne2000_get_eth_addr;

	/* manually set the port for now */
	dp->port = 0x300;

	/* Setup the default internal memory configuration for the NS8390 */
	/* TODO: This code assumes that we only talk to a 16bit card */
	dp->tx_start_page = 0x40;
	dp->stop_page     = 0x80;
	dp->rx_start_page = dp->tx_start_page + NUM_TX_PAGES;



	/* Setup ColdFire interrupts */
	sc = rtems_interrupt_catch(	ns8390_interrupt_handler,
								NIC_IRQ_VECTOR,
								&old_handler );
	if (sc != RTEMS_SUCCESSFUL)
		rtems_panic ("Can't attach 68HC901(NE2000) interrupt handler: %s\n",
  						rtems_status_text (sc));
#if 0
	sim->icr4 =	MCF5206_SIM_ICR_IL(NIC_IRQ_LEVEL) |
				MCF5206_SIM_ICR_IP(NIC_IRQ_PRIORITY);
	sim->imr = sim->imr & (~MCF5206_SIM_IMR_EINT4);	/* Enable 68HC901 interrupts */
#else
	sim->avcr = sim->avcr | MCF5307_SIM_AVCR_AVEC3;
	sim->imr = sim->imr & (~MCF5307_SIM_IMR_EINT3);
#endif


	status = ns8390_driver_attach( config, dp );


	if (debug_level >= DEBUG_STD)
		printf( "ne2000_driver_attach() - end\n");

	return status;
};


/* Where the hell does this go!?!? */

#if 0

/* handy variable decs for this stuff */


	/*
	 * Set up NS8390 hardware
	 */

	ns8390_init( dp );



	/* Add the network interface to the system and start the chip */
	ns8390_start( dp );



	return 0;
}

#endif